"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from typing_extensions import Protocol

from cfnlint.template.transforms._types import TransformResult


class Transformer(Protocol):
    def transform(self, cfn: Any) -> TransformResult: ...  # noqa: E704



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Parameters are configured correctly"""

    id = "E2001"
    shortdesc = "Parameters have appropriate properties"
    description = "Making sure the parameters are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"
    tags = ["parameters"]

    valid_keys = {
        "AllowedPattern": {"Type": "String"},
        "AllowedValues": {
            "Type": "List",
            "ItemType": "String",
        },
        "ConstraintDescription": {"Type": "String"},
        "Default": {"Type": "String"},
        "Description": {"Type": "String"},
        "MaxLength": {
            "Type": "Integer",
            "ValidForTypes": [
                "String",
                "AWS::EC2::AvailabilityZone::Name",
                "AWS::EC2::Image::Id",
                "AWS::EC2::Instance::Id",
                "AWS::EC2::KeyPair::KeyName",
                "AWS::EC2::SecurityGroup::GroupName",
                "AWS::EC2::SecurityGroup::Id",
                "AWS::EC2::Subnet::Id",
                "AWS::EC2::Volume::Id",
                "AWS::EC2::VPC::Id",
                "AWS::Route53::HostedZone::Id",
            ],
        },
        "MaxValue": {"Type": "Integer", "ValidForTypes": ["Number"]},
        "MinLength": {
            "Type": "Integer",
            "ValidForTypes": [
                "String",
                "AWS::EC2::AvailabilityZone::Name",
                "AWS::EC2::Image::Id",
                "AWS::EC2::Instance::Id",
                "AWS::EC2::KeyPair::KeyName",
                "AWS::EC2::SecurityGroup::GroupName",
                "AWS::EC2::SecurityGroup::Id",
                "AWS::EC2::Subnet::Id",
                "AWS::EC2::Volume::Id",
                "AWS::EC2::VPC::Id",
                "AWS::Route53::HostedZone::Id",
            ],
        },
        "MinValue": {"Type": "Integer", "ValidForTypes": ["Number"]},
        "NoEcho": {"Type": "Boolean"},
        "Type": {"Type": "String"},
    }

    required_keys = ["Type"]

    def check_type(self, value, path, props):
        """Check the type and handle recursion with lists"""
        results = []
        prop_type = props.get("Type")
        if value is None:
            message = (
                f'Property {"/".join(map(str, path))} should be of type {prop_type}'
            )
            results.append(RuleMatch(path, message))
            return results
        try:
            if prop_type in ["List"]:
                if isinstance(value, list):
                    for i, item in enumerate(value):
                        results.extend(
                            self.check_type(
                                item, path[:] + [i], {"Type": props.get("ItemType")}
                            )
                        )
                else:
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
            if prop_type in ["String"]:
                if isinstance(value, (dict, list)):
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
                str(value)
            elif prop_type in ["Boolean"]:
                if not isinstance(value, bool):
                    if value not in ["True", "true", "False", "false"]:
                        message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                        results.append(RuleMatch(path, message))
            elif prop_type in ["Integer"]:
                if isinstance(value, bool):
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
                else:  # has to be a Double
                    int(value)
        except Exception:  # pylint: disable=W0703
            message = (
                f'Property {"/".join(map(str, path))} should be of type {prop_type}'
            )
            results.append(
                RuleMatch(
                    path,
                    message,
                )
            )

        return results

    def match(self, cfn):
        matches = []

        for paramname, paramvalue in cfn.get_parameters().items():
            if isinstance(paramvalue, dict):
                for propname, propvalue in paramvalue.items():
                    if propname not in self.valid_keys:
                        message = "Parameter {0} has invalid property {1}"
                        matches.append(
                            RuleMatch(
                                ["Parameters", paramname, propname],
                                message.format(paramname, propname),
                            )
                        )
                    else:
                        props = self.valid_keys.get(propname)
                        prop_path = ["Parameters", paramname, propname]
                        matches.extend(self.check_type(propvalue, prop_path, props))
                        # Check that the property is needed for the current type
                        valid_for = props.get("ValidForTypes")
                        if valid_for is not None:
                            if paramvalue.get("Type") not in valid_for:
                                message = "Parameter {0} has property {1} which is only valid for {2}"
                                matches.append(
                                    RuleMatch(
                                        ["Parameters", paramname, propname],
                                        message.format(paramname, propname, valid_for),
                                    )
                                )

                for reqname in self.required_keys:
                    if reqname not in paramvalue.keys():
                        message = "Parameter {0} is missing required property {1}"
                        matches.append(
                            RuleMatch(
                                ["Parameters", paramname],
                                message.format(paramname, reqname),
                            )
                        )
            else:
                message = "Parameter {0} is not an object"
                matches.append(
                    RuleMatch(
                        ["Parameters", paramname], message.format(paramname, reqname)
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Configuration(CloudFormationLintRule):
    """Check Update Policy Configuration"""

    id = "E3016"
    shortdesc = "Check the configuration of a resources UpdatePolicy"
    description = "Make sure a resources UpdatePolicy is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-updatepolicy.html"
    tags = ["resources", "updatepolicy"]
    valid_attributes = {
        "sub": {
            "AutoScalingReplacingUpdate": {"WillReplace": {"PrimitiveType": "Boolean"}},
            "AutoScalingRollingUpdate": {
                "MaxBatchSize": {"PrimitiveType": "Integer"},
                "MinInstancesInService": {"PrimitiveType": "Integer"},
                "MinSuccessfulInstancesPercent": {"PrimitiveType": "Integer"},
                "PauseTime": {"PrimitiveType": "String"},
                "SuspendProcesses": {
                    "PrimitiveType": "List",
                    "ValidValues": [
                        "Launch",
                        "Terminate",
                        "HealthCheck",
                        "ReplaceUnhealthy",
                        "AZRebalance",
                        "AlarmNotification",
                        "ScheduledActions",
                        "AddToLoadBalancer",
                        "InstanceRefresh",
                    ],
                },
                "WaitOnResourceSignals": {"PrimitiveType": "Boolean"},
            },
            "AutoScalingScheduledAction": {
                "IgnoreUnmodifiedGroupSizeProperties": {"PrimitiveType": "Boolean"}
            },
            "CodeDeployLambdaAliasUpdate": {
                "AfterAllowTrafficHook": {"PrimitiveType": "String"},
                "ApplicationName": {"PrimitiveType": "String"},
                "BeforeAllowTrafficHook": {"PrimitiveType": "String"},
                "DeploymentGroupName": {"PrimitiveType": "String"},
            },
        },
        "main": {
            "AutoScalingReplacingUpdate": {
                "Type": "AutoScalingReplacingUpdate",
                "ResourceTypes": ["AWS::AutoScaling::AutoScalingGroup"],
            },
            "AutoScalingRollingUpdate": {
                "Type": "AutoScalingRollingUpdate",
                "ResourceTypes": ["AWS::AutoScaling::AutoScalingGroup"],
            },
            "AutoScalingScheduledAction": {
                "Type": "AutoScalingScheduledAction",
                "ResourceTypes": ["AWS::AutoScaling::AutoScalingGroup"],
            },
            "CodeDeployLambdaAliasUpdate": {
                "Type": "CodeDeployLambdaAliasUpdate",
                "ResourceTypes": ["AWS::Lambda::Alias"],
            },
            "EnableVersionUpgrade": {
                "PrimitiveType": "Boolean",
                "ResourceTypes": [
                    "AWS::Elasticsearch::Domain",
                    "AWS::OpenSearchService::Domain",
                ],
            },
            "UseOnlineResharding": {
                "PrimitiveType": "Boolean",
                "ResourceTypes": ["AWS::ElastiCache::ReplicationGroup"],
            },
        },
        "primitive_types": {
            "String": (str),
            "Integer": (int),
            "Boolean": (bool),
            "List": list,
        },
    }

    def check_value(self, value, path, **kwargs):
        """Check a primitive value"""
        matches = []

        prim_type = kwargs.get("primitive_type")
        if prim_type == "List":
            valid_values = kwargs.get("valid_values")
            if valid_values:
                if value not in valid_values:
                    message = "Allowed values for {0} are ({1})"
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(
                                kwargs.get("key_name"),
                                ", ".join(map(str, valid_values)),
                            ),
                        )
                    )
        else:
            default_message = (
                f'Value for {kwargs.get("key_name")} must be of type {prim_type}'
            )
            if not isinstance(
                value, self.valid_attributes.get("primitive_types").get(prim_type)
            ):
                try:
                    if prim_type in ["String"]:
                        str(value)
                    elif prim_type in ["Boolean"]:
                        if value not in ["True", "true", "False", "false"]:
                            matches.append(RuleMatch(path, default_message))
                    else:  # has to be integer
                        if isinstance(value, bool):
                            matches.append(RuleMatch(path, default_message))
                        if isinstance(value, float):
                            if not value.is_integer():
                                matches.append(RuleMatch(path, default_message))
                        else:
                            int(value)
                except Exception:  # pylint: disable=W0703
                    matches.append(RuleMatch(path, default_message))

        return matches

    def check_attributes(self, cfn, properties, spec_type, path):
        """
        Check the properties against the spec
        """
        matches = []
        spec = self.valid_attributes.get("sub").get(spec_type)
        if isinstance(properties, dict):
            for p_value_safe, p_path_safe in properties.items_safe(path):
                for p_name, p_value in p_value_safe.items():
                    if p_name in spec:
                        up_type_spec = spec.get(p_name)
                        if "Type" in up_type_spec:
                            matches.extend(
                                self.check_attributes(
                                    cfn,
                                    p_value,
                                    up_type_spec.get("Type"),
                                    p_path_safe[:] + [p_name],
                                )
                            )
                        else:
                            matches.extend(
                                cfn.check_value(
                                    obj={p_name: p_value},
                                    key=p_name,
                                    path=p_path_safe[:],
                                    check_value=self.check_value,
                                    valid_values=up_type_spec.get("ValidValues", []),
                                    primitive_type=up_type_spec.get("PrimitiveType"),
                                    key_name=p_name,
                                )
                            )
                    else:
                        message = "UpdatePolicy doesn't support type {0}"
                        matches.append(
                            RuleMatch(path[:] + [p_name], message.format(p_name))
                        )
        else:
            message = "{0} should be an object"
            matches.append(RuleMatch(path, message.format(path[-1])))

        return matches

    def check_update_policy(self, cfn, update_policy, path, resource_type):
        """Check an update policy"""
        matches = []
        for up_type, up_value in update_policy.items():
            up_path = path[:] + [up_type]
            up_type_spec = self.valid_attributes.get("main").get(up_type)
            if up_type_spec:
                if resource_type not in up_type_spec.get("ResourceTypes"):
                    message = "UpdatePolicy only supports this type for resources of type ({0})"
                    matches.append(
                        RuleMatch(
                            up_path,
                            message.format(
                                ", ".join(map(str, up_type_spec.get("ResourceTypes")))
                            ),
                        )
                    )
                if "Type" in up_type_spec:
                    matches.extend(
                        self.check_attributes(
                            cfn, up_value, up_type_spec.get("Type"), up_path[:]
                        )
                    )

                else:
                    matches.extend(
                        cfn.check_value(
                            obj={up_type: up_value},
                            key=up_type,
                            path=up_path[:],
                            check_value=self.check_value,
                            valid_values=up_type_spec.get("ValidValues", []),
                            primitive_type=up_type_spec.get("PrimitiveType"),
                            key_name=up_type,
                        )
                    )
            else:
                message = "UpdatePolicy doesn't support type {0}"
                matches.append(RuleMatch(up_path, message.format(up_type)))

        return matches

    def match(self, cfn):
        matches = []

        for r_name, r_values in cfn.get_resources().items():
            update_policy = r_values.get("UpdatePolicy")
            if update_policy is None:
                continue
            resource_type = r_values.get("Type")
            if isinstance(update_policy, dict):
                path = ["Resources", r_name, "UpdatePolicy"]
                for up_value_safe, up_path_safe in update_policy.items_safe(path):
                    matches.extend(
                        self.check_update_policy(
                            cfn, up_value_safe, up_path_safe, resource_type
                        )
                    )
            else:
                message = "UpdatePolicy should be an object"
                matches.append(
                    RuleMatch(["Resources", r_name, "UpdatePolicy"], message.format())
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from datetime import date

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class PolicyVersion(CloudFormationLintRule):
    """Check if IAM Policy Version is correct"""

    id = "W2511"
    shortdesc = "Check IAM Resource Policies syntax"
    description = (
        "See if the elements inside an IAM Resource policy are configured correctly."
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html"
    tags = ["properties", "iam"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resources_and_keys = {
            "AWS::ECR::Repository": "RepositoryPolicyText",
            "AWS::Elasticsearch::Domain": "AccessPolicies",
            "AWS::OpenSearchService::Domain": "AccessPolicies",
            "AWS::KMS::Key": "KeyPolicy",
            "AWS::S3::BucketPolicy": "PolicyDocument",
            "AWS::SNS::TopicPolicy": "PolicyDocument",
            "AWS::SQS::QueuePolicy": "PolicyDocument",
        }
        self.idp_and_keys = {
            "AWS::IAM::Group": "Policies",
            "AWS::IAM::ManagedPolicy": "PolicyDocument",
            "AWS::IAM::Policy": "PolicyDocument",
            "AWS::IAM::Role": "Policies",
            "AWS::IAM::User": "Policies",
        }
        for resource_type in self.resources_and_keys:
            self.resource_property_types.append(resource_type)
        for resource_type in self.idp_and_keys:
            self.resource_property_types.append(resource_type)

    def check_policy_document(self, value, path):
        """Check policy document"""
        matches = []

        if not isinstance(value, dict):
            return matches

        for e_v, e_p in value.items_safe(path[:]):
            for p_vs, p_p in e_v.items_safe(e_p[:]):
                version = p_vs.get("Version")
                if version:
                    if version in ["2008-10-17", date(2008, 10, 17)]:
                        message = (
                            "IAM Policy Version should be updated to '2012-10-17'."
                        )
                        matches.append(RuleMatch(p_p + ["Version"], message))
        return matches

    def match_resource_properties(self, properties, resourcetype, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        key = None
        if resourcetype in self.resources_and_keys:
            key = self.resources_and_keys.get(resourcetype)
        else:
            key = self.idp_and_keys.get(resourcetype)

        if key == "Policies":
            for index, policy in enumerate(properties.get(key, [])):
                matches.extend(
                    cfn.check_value(
                        obj=policy,
                        key="PolicyDocument",
                        path=path[:] + ["Policies", index],
                        check_value=self.check_policy_document,
                    )
                )
        else:
            matches.extend(
                cfn.check_value(
                    obj=properties,
                    key=key,
                    path=path[:],
                    check_value=self.check_policy_document,
                )
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Conditions are configured correctly"""

    id = "E8001"
    shortdesc = "Conditions have appropriate properties"
    description = "Check if Conditions are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/conditions-section-structure.html"
    tags = ["conditions"]

    condition_keys = [
        "Condition",
        "Fn::And",
        "Fn::Equals",
        "Fn::Not",
        "Fn::Or",
    ]

    def match(self, cfn):
        matches = []

        conditions = cfn.template.get("Conditions", {})
        if conditions:
            for condname, condobj in conditions.items():
                if not isinstance(condobj, dict):
                    message = "Condition {0} has invalid property"
                    matches.append(
                        RuleMatch(["Conditions", condname], message.format(condname))
                    )
                else:
                    if len(condobj) != 1:
                        message = "Condition {0} has too many intrinsic conditions"
                        matches.append(
                            RuleMatch(
                                ["Conditions", condname], message.format(condname)
                            )
                        )
                    else:
                        for k, _ in condobj.items():
                            if k not in self.condition_keys:
                                message = "Condition {0} has invalid property {1}"
                                matches.append(
                                    RuleMatch(
                                        ["Conditions", condname] + [k],
                                        message.format(condname, k),
                                    )
                                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RefWithPath(CloudFormationLintRule):
    """Check if IAM Policy Version is correct"""

    id = "E3050"
    shortdesc = "Check if REFing to a IAM resource with path set"
    description = (
        "Some resources don't support looking up the IAM resource by name. "
        "This check validates when a REF is being used and the Path is not '/'"
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html"
    tags = ["properties", "iam"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resources_and_keys = {
            "AWS::CodeBuild::Project": "ServiceRole",
        }

        for resource_type in self.resources_and_keys:
            self.resource_property_types.append(resource_type)

    def check_ref(self, value, parameters, resources, path):  # pylint: disable=W0613
        """Check Ref"""
        matches = []

        iam_path = resources.get(value, {}).get("Properties", {}).get("Path")
        if not iam_path:
            return matches

        if iam_path != "/":
            message = "When using a Ref to IAM resource the Path must be '/'.  Switch to GetAtt if the Path has to be '{}'."
            matches.append(RuleMatch(path, message.format(iam_path)))

        return matches

    def match_resource_properties(self, properties, resourcetype, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        key = None
        if resourcetype in self.resources_and_keys:
            key = self.resources_and_keys.get(resourcetype)

        matches.extend(cfn.check_value(properties, key, path, check_ref=self.check_ref))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DefaultRef(CloudFormationLintRule):
    """Check if Parameter defaults don't use Refs"""

    id = "E2014"
    shortdesc = "Default value cannot use Refs"
    description = "Check if Refs are not used in Parameter Defaults"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"
    tags = ["parameters", "ref"]

    def match(self, cfn):
        matches = []

        ref_objs = cfn.search_deep_keys("Ref")

        # Filter out the Parameters
        parameter_refs = [x for x in ref_objs if x[0] == "Parameters"]

        for parameter_ref in parameter_refs:
            message = "Invalid value ({}), Ref cannot be used in Parameters"
            matches.append(RuleMatch(parameter_ref, message.format(parameter_ref[-1])))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging

from cfnlint.languageExtensions import LanguageExtensions
from cfnlint.rules import CloudFormationLintRule

LOGGER = logging.getLogger("cfnlint")


class ForEach(CloudFormationLintRule):
    id = "E1032"
    shortdesc = "Validates ForEach functions"
    description = "Validates that ForEach parameters have a valid configuration"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-getatt.html"
    tags = ["functions", "foreach"]

    # pylint: disable=unused-argument
    def match(self, cfn):
        matches = []
        intrinsic_function = "Fn::ForEach"
        for_eaches = cfn.transform_pre["Fn::ForEach"]

        for for_each in for_eaches:
            has_language_extensions_transform = cfn.has_language_extensions_transform()

            LanguageExtensions().validate_transform_is_declared(
                has_language_extensions_transform,
                matches,
                for_each[:-1],
                intrinsic_function,
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(CloudFormationLintRule):
    """Check for UpdateReplacePolicy / DeletionPolicy"""

    id = "I3011"
    shortdesc = "Check stateful resources have a set UpdateReplacePolicy/DeletionPolicy"
    description = (
        "The default action when replacing/removing a resource is to "
        "delete it. This check requires you to explicitly set policies"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"
    tags = ["resources", "updatereplacepolicy", "deletionpolicy"]

    def __init__(self):
        """Init"""
        super().__init__()

        spec = cfnlint.helpers.load_resource(AdditionalSpecs, "StatefulResources.json")

        self.likely_stateful_resource_types = [
            resource_type
            for resource_type, descr in spec["ResourceTypes"].items()
            # Resources that won't be deleted if they're not empty (ex: S3)
            # don't need to be checked for policies, as chance of mistakes are low.
            if not descr.get("DeleteRequiresEmptyResource", False)
        ]

    def match(self, cfn):
        """Check for UpdateReplacePolicy / DeletionPolicy"""
        matches = []

        resources = cfn.get_resources()
        for r_name, r_values in resources.items():
            if r_values.get("Type") in self.likely_stateful_resource_types:
                if not r_values.get("DeletionPolicy") or not r_values.get(
                    "UpdateReplacePolicy"
                ):
                    path = ["Resources", r_name]
                    message = f'The default action when replacing/removing a resource is to delete it. Set explicit values for UpdateReplacePolicy / DeletionPolicy on potentially stateful resource: {"/".join(path)}'
                    matches.append(RuleMatch(path, message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AllowedPattern(CloudFormationLintRule):
    """Check if parameters have a valid value"""

    id = "W2031"
    shortdesc = "Check if parameters have a valid value based on an allowed pattern"
    description = "Check if parameters have a valid value in a pattern. The Parameter's allowed pattern is based on the usages in property (Ref)"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedpattern"
    tags = ["parameters", "resources", "property", "allowed pattern"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def check_value_ref(self, value, path, **kwargs):
        """Check Ref"""
        matches = []

        cfn = kwargs.get("cfn")
        if "Fn::If" in path:
            self.logger.debug(
                "Not able to guarentee that the default value hasn't been conditioned out"
            )
            return matches
        if path[0] == "Resources" and "Condition" in cfn.template.get(path[0], {}).get(
            path[1]
        ):
            self.logger.debug(
                "Not able to guarentee that the default value "
                "hasn't been conditioned out"
            )
            return matches

        allowed_pattern = kwargs.get("value_specs", {}).get("AllowedPattern", {})
        allowed_pattern_regex = kwargs.get("value_specs", {}).get(
            "AllowedPatternRegex", {}
        )
        allowed_pattern_description = kwargs.get("value_specs", {}).get(
            "AllowedPatternDescription", {}
        )

        if allowed_pattern_regex:
            if value in cfn.template.get("Parameters", {}):
                param = cfn.template.get("Parameters").get(value, {})
                parameter_values = param.get("AllowedValues")
                default_value = param.get("Default")
                parameter_type = param.get("Type")
                if isinstance(parameter_type, str):
                    if (
                        (not parameter_type.startswith("List<"))
                        and (
                            not parameter_type.startswith("AWS::SSM::Parameter::Value<")
                        )
                        and parameter_type not in ["CommaDelimitedList", "List<String>"]
                    ):
                        # Check Allowed Values
                        if parameter_values:
                            for index, allowed_value in enumerate(parameter_values):
                                if not re.match(
                                    allowed_pattern_regex, str(allowed_value)
                                ):
                                    param_path = [
                                        "Parameters",
                                        value,
                                        "AllowedValues",
                                        index,
                                    ]
                                    description = (
                                        allowed_pattern_description
                                        or f"Valid values must match pattern {allowed_pattern}"
                                    )
                                    message = "You must specify a valid allowed value for {0} ({1}). {2}"
                                    matches.append(
                                        RuleMatch(
                                            param_path,
                                            message.format(
                                                value, allowed_value, description
                                            ),
                                        )
                                    )
                        if default_value:
                            # Check Default, only if no allowed Values are specified in the parameter (that's covered by E2015)
                            if not re.match(allowed_pattern_regex, str(default_value)):
                                param_path = ["Parameters", value, "Default"]
                                description = (
                                    allowed_pattern_description
                                    or f"Valid values must match pattern {allowed_pattern}"
                                )
                                message = "You must specify a valid Default value for {0} ({1}). {2}"
                                matches.append(
                                    RuleMatch(
                                        param_path,
                                        message.format(
                                            value, default_value, description
                                        ),
                                    )
                                )

        return matches

    def check(self, cfn, properties, value_specs, property_specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in value_specs:
                    value = value_specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ValueType", "")
                        property_type = (
                            property_specs.get("Properties").get(prop).get("Type")
                        )
                        value_specs = (
                            RESOURCE_SPECS.get(cfn.regions[0])
                            .get("ValueTypes")
                            .get(value_type, {})
                        )
                        if value_specs == "CACHED":
                            value_specs = (
                                RESOURCE_SPECS.get("us-east-1")
                                .get("ValueTypes")
                                .get(value_type, {})
                            )
                        matches.extend(
                            cfn.check_value(
                                p_value,
                                prop,
                                p_path,
                                check_ref=self.check_value_ref,
                                value_specs=value_specs,
                                cfn=cfn,
                                property_type=property_type,
                                property_name=prop,
                            )
                        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        property_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("PropertyTypes").get(property_type)
        )
        matches.extend(self.check(cfn, properties, specs, property_specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        resource_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("ResourceTypes").get(resource_type)
        )
        matches.extend(self.check(cfn, properties, specs, resource_specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.languageExtensions import LanguageExtensions
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Length(CloudFormationLintRule):
    """Check if Length values are correct"""

    id = "E1030"
    shortdesc = "Length validation of parameters"
    description = "Making sure Fn::Length is configured correctly"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-length.html"
    tags = ["functions", "length"]

    supported_functions = ["Fn::Split", "Fn::FindInMap"]

    def match(self, cfn):
        has_language_extensions_transform = cfn.has_language_extensions_transform()
        intrinsic_function = "Fn::Length"
        matches = []

        fn_length_objects = cfn.search_deep_keys(intrinsic_function)
        for fn_length_object in fn_length_objects:
            tree = fn_length_object[:-1]
            LanguageExtensions().validate_transform_is_declared(
                has_language_extensions_transform,
                matches,
                tree,
                intrinsic_function,
            )
            self.validate_type(fn_length_object, matches, tree)
            self.validate_ref(fn_length_object, matches, tree, cfn)
        return matches

    def validate_ref(self, fn_length_object, matches, tree, cfn):
        fn_length_value = fn_length_object[-1]
        if isinstance(fn_length_value, dict):
            if len(fn_length_value.keys()) != 1 or (
                list(fn_length_value.keys())[0]
                not in ["Ref"] + self.supported_functions
            ):
                self.addMatch(
                    matches,
                    tree,
                    f"Fn::Length expects either an array, a Ref to an array or {', '.join(self.supported_functions)}, "
                    "but found unexpected object under Fn::Length at {0}",
                )
                return

            if "Ref" in fn_length_value:
                if fn_length_value["Ref"] not in cfn.get_parameter_names():
                    self.addMatch(
                        matches,
                        tree,
                        "Fn::Length can only reference list parameters at {0}",
                    )
                else:
                    referenced_parameter = cfn.get_parameters().get(
                        fn_length_value["Ref"]
                    )
                    parameter_type = referenced_parameter.get("Type")
                    if "List" not in parameter_type:
                        self.addMatch(
                            matches,
                            tree,
                            "Fn::Length can only reference list parameters at {0}",
                        )

    def addMatch(self, matches, tree, message):
        matches.append(RuleMatch(tree[:], message.format("/".join(map(str, tree)))))

    def validate_type(self, fn_length_object, matches, tree):
        fn_length_value = fn_length_object[-1]
        if not isinstance(fn_length_value, dict) and not isinstance(
            fn_length_value, list
        ):
            message = "Fn::Length needs a list or a reference to a list at {0}"
            matches.append(RuleMatch(tree[:], message.format("/".join(map(str, tree)))))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""


class Mark:
    """Mark of line and column"""

    line = 1
    column = 1

    def __init__(self, line, column):
        self.line = line
        self.column = column



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ZipPackageRequiredProperties(CloudFormationLintRule):
    id = "W2533"
    shortdesc = (
        "Check required properties for Lambda if the deployment package is a .zip file"
    )
    description = (
        "When the package type is Zip, "
        "you must also specify the `handler` and `runtime` properties."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-lambda-function.html"
    tags = ["resources", "lambda"]

    def match(self, cfn):
        matches = []
        required_properties = [
            "Handler",
            "Runtime",
        ]  # required if package is a .zip file

        resources = cfn.get_resources(["AWS::Lambda::Function"])

        for resource_name, resource in resources.items():
            properties = resource.get("Properties")
            if not isinstance(properties, dict):
                continue

            for scenario in cfn.get_object_without_conditions(
                properties, ["PackageType", "Code", "Handler", "Runtime"]
            ):
                props = scenario.get("Object")
                path = ["Resources", resource_name, "Properties"]

                # check is zip deployment
                is_zip_deployment = True
                code = props.get("Code")

                if props.get("PackageType") == "Zip":
                    path.append("PackageType")
                elif isinstance(code, dict) and (
                    code.get("ZipFile") or code.get("S3Key")
                ):
                    path.append("Code")
                else:
                    is_zip_deployment = False

                if not is_zip_deployment:
                    continue

                # check required properties for zip deployment
                missing_properties = []
                for p in required_properties:
                    if props.get(p) is None:
                        missing_properties.append(p)

                if len(missing_properties) > 0:
                    message = "Properties {0} missing for zip file deployment at {1}"
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(
                                missing_properties,
                                "/".join(
                                    map(str, ["Resources", resource_name, "Properties"])
                                ),
                            ),
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Ref(CloudFormationLintRule):
    """Check if Ref value is a string"""

    id = "E1020"
    shortdesc = "Ref validation of value"
    description = (
        "Making sure the Ref has a String value (no other functions are supported)"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["functions", "ref"]

    def match(self, cfn):
        matches = []

        ref_objs = cfn.search_deep_keys("Ref")
        for ref_obj in ref_objs:
            value = ref_obj[-1]
            if not isinstance(value, (str)):
                message = "Ref can only be a string for {0}"
                matches.append(
                    RuleMatch(
                        ref_obj[:-1], message.format("/".join(map(str, ref_obj[:-1])))
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ValuePrimitiveType(CloudFormationLintRule):
    """Check if Resource PrimitiveTypes are correct"""

    id = "E3012"
    shortdesc = "Check resource properties values"
    description = "Checks resource property values with Primitive Types for values that match those types."
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#valueprimitivetype"
    tags = ["resources"]

    strict_exceptions = {
        "AWS::CloudFormation::Stack": ["Parameters"],
        "AWS::Lambda::Function.Environment": ["Variables"],
    }

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_specs = []
        self.property_specs = []
        self.config_definition = {
            "strict": {"default": False, "type": "boolean"}
        }  # Strict mode is set to false by default
        self.configure()

    def initialize(self, cfn):
        """Initialize the rule"""
        specs = cfnlint.helpers.RESOURCE_SPECS.get(cfn.regions[0])
        self.property_specs = specs.get("PropertyTypes")
        self.resource_specs = specs.get("ResourceTypes")
        for resource_spec in self.resource_specs:
            self.resource_property_types.append(resource_spec)
        for property_spec in self.property_specs:
            self.resource_sub_property_types.append(property_spec)

    def _value_check(self, value, path, item_type, strict_check, extra_args):
        """Checks non strict"""
        matches = []
        if not strict_check:
            try:
                if item_type in ["String"]:
                    str(value)
                elif item_type in ["Boolean"]:
                    if value not in ["True", "true", "False", "false"]:
                        message = f'Property {"/".join(map(str, path))} should be of type {item_type}'
                        matches.append(RuleMatch(path, message, **extra_args))
                elif item_type in ["Integer", "Long", "Double"]:
                    if isinstance(value, bool):
                        message = f'Property {"/".join(map(str, path))} should be of type {item_type}'
                        matches.append(RuleMatch(path, message, **extra_args))
                    elif item_type in ["Integer"]:
                        int(value)
                    elif item_type in ["Long"]:
                        # Some times python will strip the decimals when doing a conversion
                        if isinstance(value, float):
                            message = f'Property {"/".join(map(str, path))} should be of type {item_type}'
                            matches.append(RuleMatch(path, message, **extra_args))
                        int(value)
                    else:  # has to be a Double
                        float(value)
            except Exception:  # pylint: disable=W0703
                message = (
                    f'Property {"/".join(map(str, path))} should be of type {item_type}'
                )
                matches.append(RuleMatch(path, message, **extra_args))
        else:
            message = (
                f'Property {"/".join(map(str, path))} should be of type {item_type}'
            )
            matches.append(RuleMatch(path, message, **extra_args))

        return matches

    def check_primitive_type(self, value, item_type, path, strict_check):
        """Chec item type"""
        matches = []
        if isinstance(value, (dict, list)) and item_type == "Json":
            return matches
        if item_type in ["String"]:
            if not isinstance(value, (str)):
                extra_args = {
                    "actual_type": type(value).__name__,
                    "expected_type": str.__name__,
                }
                matches.extend(
                    self._value_check(value, path, item_type, strict_check, extra_args)
                )
        elif item_type in ["Boolean"]:
            if not isinstance(value, (bool)):
                extra_args = {
                    "actual_type": type(value).__name__,
                    "expected_type": bool.__name__,
                }
                matches.extend(
                    self._value_check(value, path, item_type, strict_check, extra_args)
                )
        elif item_type in ["Double"]:
            if not isinstance(value, (float, int)):
                extra_args = {
                    "actual_type": type(value).__name__,
                    "expected_type": [float.__name__, int.__name__],
                }
                matches.extend(
                    self._value_check(value, path, item_type, strict_check, extra_args)
                )
        elif item_type in ["Integer"]:
            if not isinstance(value, (int)):
                extra_args = {
                    "actual_type": type(value).__name__,
                    "expected_type": int.__name__,
                }
                matches.extend(
                    self._value_check(value, path, item_type, strict_check, extra_args)
                )
        elif item_type in ["Long"]:
            integer_types = (int,)
            if not isinstance(value, integer_types):
                extra_args = {
                    "actual_type": type(value).__name__,
                    "expected_type": " or ".join([x.__name__ for x in integer_types]),
                }
                matches.extend(
                    self._value_check(value, path, item_type, strict_check, extra_args)
                )
        elif isinstance(value, list):
            message = (
                f'Property should be of type {item_type} at {"/".join(map(str, path))}'
            )
            extra_args = {
                "actual_type": type(value).__name__,
                "expected_type": list.__name__,
            }
            matches.append(RuleMatch(path, message, **extra_args))

        return matches

    def check_value(self, value, path, **kwargs):
        """Check Value"""
        matches = []
        primitive_type = kwargs.get("primitive_type", {})
        item_type = kwargs.get("item_type", {})
        strict_check = kwargs.get("non_strict", self.config["strict"])

        if value is None:
            message = f'Property value cannot be null {"/".join(map(str, path))}'
            matches.append(RuleMatch(path, message))
        elif item_type in ["Map"]:
            if isinstance(value, dict):
                for map_key, map_value in value.items():
                    if not isinstance(map_value, dict):
                        matches.extend(
                            self.check_primitive_type(
                                map_value,
                                primitive_type,
                                path + [map_key],
                                strict_check,
                            )
                        )
                    else:
                        # types that represent a singular value (not json)
                        cfnlint.helpers.FUNCTIONS_SINGLE.sort()
                        if primitive_type in ["String", "Boolean", "Integer", "Double"]:
                            if len(map_value) != 1:
                                matches.append(
                                    RuleMatch(
                                        path,
                                        f'Use a valid function [{", ".join(cfnlint.helpers.FUNCTIONS_SINGLE)}] when providing a value of type [{primitive_type}]',
                                    )
                                )
                            else:
                                for k in map_value.keys():
                                    if k not in cfnlint.helpers.FUNCTIONS_SINGLE:
                                        matches.append(
                                            RuleMatch(
                                                path,
                                                f'Use a valid function [{", ".join(cfnlint.helpers.FUNCTIONS_SINGLE)}] when providing a value of type [{primitive_type}]',
                                            )
                                        )
        else:
            # some properties support primitive types and objects
            # skip in the case it could be an object and the value is a object
            if (item_type or primitive_type) and isinstance(value, dict):
                return matches
            matches.extend(
                self.check_primitive_type(value, primitive_type, path, strict_check)
            )

        return matches

    def _check_map(self, m, path):
        matches = []
        if isinstance(m, dict):
            if len(m) == 1:
                for k, v in m.items():
                    if k == "Fn::If":
                        if isinstance(v, list) and len(v) == 3:
                            matches.extend(
                                self._check_map(v[1], path[:] + ["Fn::If", 1])
                            )
                            matches.extend(
                                self._check_map(v[2], path[:] + ["Fn::If", 2])
                            )
        else:
            matches.append(
                RuleMatch(
                    path,
                    "Map must be an object of key-value pairs",
                )
            )
        return matches

    def check(self, cfn, properties, specs, spec_type, path):
        """Check itself"""
        matches = []

        for prop in properties:
            if prop in specs:
                primitive_type = specs.get(prop).get("PrimitiveType")
                if not primitive_type:
                    primitive_type = specs.get(prop).get("PrimitiveItemType")
                if specs.get(prop).get("Type") == "Map":
                    matches.extend(
                        self._check_map(properties.get(prop), path[:] + [prop])
                    )
                if specs.get(prop).get("Type") in ["List", "Map"]:
                    item_type = specs.get(prop).get("Type")
                else:
                    item_type = None
                if primitive_type:
                    strict_check = self.config["strict"]
                    if spec_type in self.strict_exceptions:
                        if prop in self.strict_exceptions[spec_type]:
                            strict_check = False
                    matches.extend(
                        cfn.check_value(
                            properties,
                            prop,
                            path,
                            check_value=self.check_value,
                            primitive_type=primitive_type,
                            item_type=item_type,
                            non_strict=strict_check,
                            pass_if_null=True,
                        )
                    )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        if self.property_specs.get(property_type, {}).get("Properties"):
            property_specs = self.property_specs.get(property_type, {}).get(
                "Properties", {}
            )
            matches.extend(
                self.check(cfn, properties, property_specs, property_type, path)
            )

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        resource_specs = self.resource_specs.get(resource_type, {}).get(
            "Properties", {}
        )
        matches.extend(self.check(cfn, properties, resource_specs, resource_type, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.decode.mark import Mark
from cfnlint.decode.node import dict_node, list_node, str_node


def convert_dict(template, start_mark=Mark(0, 0), end_mark=Mark(0, 0)):
    """Convert dict to template"""
    if isinstance(template, dict):
        if not isinstance(template, dict_node):
            template = dict_node(template, start_mark, end_mark)
        for k, v in template.copy().items():
            k_start_mark = start_mark
            k_end_mark = end_mark
            if isinstance(k, str_node):
                k_start_mark = k.start_mark
                k_end_mark = k.end_mark
            new_k = str_node(k, k_start_mark, k_end_mark)
            del template[k]
            template[new_k] = convert_dict(v, k_start_mark, k_end_mark)
    elif isinstance(template, list):
        if not isinstance(template, list_node):
            template = list_node(template, start_mark, end_mark)
        for i, v in enumerate(template):
            template[i] = convert_dict(v, start_mark, end_mark)

    return template



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
import random
import string
from copy import deepcopy
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Tuple

import regex as re

from cfnlint.conditions._utils import get_hash
from cfnlint.decode.node import str_node
from cfnlint.helpers import FUNCTION_FOR_EACH
from cfnlint.template.transforms._types import TransformResult

LOGGER = logging.getLogger("cfnlint")

# initializing size of string
_N = 7

_SCALAR_TYPES = (str, int, float, bool)


class _ResolveError(Exception):
    def __init__(self, message: str, key: str) -> None:
        super().__init__(message)
        self.key = key


class _ValueError(Exception):
    def __init__(self, message: str, key: str) -> None:
        super().__init__(message)
        self.key = key


class _TypeError(Exception):
    def __init__(self, message: str, key: str) -> None:
        super().__init__(message)
        self.key = key


def language_extension(cfn: Any) -> TransformResult:
    transform = _Transform()
    try:
        return transform.transform(cfn)
    except (_ValueError, _TypeError, _ResolveError) as e:
        # pylint: disable=import-outside-toplevel
        from cfnlint.match import Match  # pylint: disable=cyclic-import
        from cfnlint.rules import TransformError  # pylint: disable=cyclic-import

        message = "Error transforming template: {0}"
        if hasattr(e.key, "start_mark"):
            sm_line = e.key.start_mark.line + 1
            sm_column = e.key.start_mark.column + 1
        else:
            sm_line = 1
            sm_column = 1
        if hasattr(e.key, "end_mark"):
            em_line = e.key.end_mark.line + 1
            em_column = e.key.end_mark.column + 1
        else:
            em_line = 1
            em_column = 1

        return [
            Match(
                sm_line,
                sm_column,
                em_line,
                em_column,
                cfn.filename,
                TransformError(),
                message.format(str(e)),
            )
        ], None
    except Exception as e:  # pylint: disable=broad-exception-caught
        # pylint: disable=import-outside-toplevel
        from cfnlint.match import Match  # pylint: disable=cyclic-import
        from cfnlint.rules import TransformError  # pylint: disable=cyclic-import

        message = "Error transforming template: {0}"
        return [
            Match(
                1,
                1,
                1,
                1,
                cfn.filename,
                TransformError(),
                message.format(str(e)),
            )
        ], None


class _Transform:
    def __init__(self) -> None:
        self._collections: MutableMapping[str, str] = {}

    def transform(self, cfn: Any) -> TransformResult:
        """Transform the template"""
        return [], self._walk(cfn.template, {}, cfn)

    # pylint: disable=too-many-return-statements
    def _walk(self, item: Any, params: MutableMapping[str, Any], cfn: Any):
        obj = deepcopy(item)
        if isinstance(obj, dict):
            # adjust keys if needed
            if params:
                for k, v in item.items():
                    _, new_k = self._replace_string_params(k, params)
                    if new_k != k:
                        del obj[k]
                        obj[new_k] = v

            for k, v in deepcopy(obj).items():
                # see if key matches Fn::ForEach
                if re.match(FUNCTION_FOR_EACH, k):
                    # only translate the foreach if its valid
                    foreach = _ForEach(k, v, self._collections)
                    # get the values will flatten the foreach
                    for collection_value in foreach.items(cfn):
                        flattened = self._walk(
                            v[2], {**params, **{v[0]: collection_value}}, cfn
                        )
                        for f_k, f_v in flattened.items():
                            if f_k not in obj:
                                obj[f_k] = f_v
                            else:
                                raise _ValueError(
                                    f"Duplicate {f_k} while doing transformation",
                                    f_k,
                                )
                    del obj[k]
                elif k == "Fn::ToJsonString":
                    # extra special handing for this as {} could be a valid value
                    return obj
                elif k == "Fn::Sub":
                    if isinstance(v, str):
                        only_string, obj[k] = self._replace_string_params(v, params)
                        if only_string:
                            return obj[k]
                    if isinstance(v, list):
                        only_string, obj[k][0] = self._replace_string_params(
                            v[0],
                            params,
                        )
                        if only_string:
                            return obj[k][0]
                        if len(v) == 2:
                            obj[k][1] = self._walk(v[1], params, cfn)
                elif k == "Fn::FindInMap":
                    try:
                        mapping = _ForEachValueFnFindInMap(get_hash(v), v)
                        map_value = mapping.value(cfn, params, True)
                        # if we get None this means its all strings but couldn't be resolved
                        # we will pass this forward
                        if map_value is None:
                            continue
                        # if we can resolve it we will return it
                        if isinstance(map_value, tuple([list]) + _SCALAR_TYPES):
                            return map_value
                    except Exception as e:  # pylint: disable=broad-exception-caught
                        # We couldn't resolve the FindInMap so we are going to leave it as it is
                        LOGGER.debug("Transform and Fn::FindInMap error: %s", {str(e)})
                elif k == "Ref":
                    if isinstance(v, str):
                        if v in params:
                            return params[v]
                    elif isinstance(v, dict):
                        r = self._walk(v, params, cfn)
                        if isinstance(r, str):
                            if r in params:
                                return params[r]
                        obj[k] = r
                else:
                    sub_value = self._walk(v, params, cfn)
                    # a sub object may be none or we have returned
                    # an empty object.  We don't want to remove empty
                    # strings "" or 0 (zeros)
                    # Remove `or sub_value == {}` for issue #2896
                    if sub_value is None:
                        del obj[k]
                    else:
                        obj[k] = sub_value
        elif isinstance(obj, list):
            for i, v in enumerate(obj):
                obj[i] = self._walk(v, params, cfn)
        return obj

    def _replace_string_params(
        self, s: str, params: Mapping[str, Any]
    ) -> Tuple[bool, str]:
        pattern = r"(\$|&){[a-zA-Z0-9\.:]+}"
        if not re.search(pattern, s):
            return (True, s)

        new_s = deepcopy(s)
        for k, v in params.items():
            new_s = re.sub(rf"\$\{{{k}\}}", v, new_s)
            new_s = re.sub(rf"\&\{{{k}\}}", re.sub("[^0-9a-zA-Z]+", "", v), new_s)

        if isinstance(s, str_node):
            new_s = str_node(new_s, s.start_mark, s.end_mark)

        return (not (bool(re.search(pattern, new_s))), new_s)


class _ForEachValue:
    def __init__(self, _hash: str, value: Any = None) -> None:
        if value:
            self._value = value

        self._hash = _hash

    @staticmethod
    def create(obj: Any) -> _ForEachValue:
        _hash = get_hash(obj)
        if isinstance(obj, _SCALAR_TYPES):
            return _ForEachValue(_hash, obj)
        if isinstance(obj, dict):
            if len(obj) != 1:
                raise _ValueError("Object must have only one key", obj)

            for k, v in obj.items():
                if k == "Ref":
                    return _ForEachValueRef(_hash, v)
                if k == "Fn::FindInMap":
                    return _ForEachValueFnFindInMap(_hash, v)

        raise _TypeError(f"Unsupported value {obj!r}", obj)

    # pylint: disable=unused-argument
    def value(
        self, cfn, params: Optional[Mapping[str, Any]] = None, only_params: bool = False
    ):
        return self._value

    @property
    def hash(self):
        return self._hash


class _FnFindInMapDefaultValue(_ForEachValue):
    def __init__(self, _hash: str, value: Any = None) -> None:
        super().__init__(_hash, value)
        if not isinstance(value, dict):
            raise _TypeError(
                "Fn::FindInMap parameter must be an object with key 'DefaultValue'",
                value,
            )
        if len(value) != 1:
            raise _ValueError(
                "Fn::FindInMap parameter only supports 'DefaultValue'", value
            )

        for k, v in value.items():
            if k != "DefaultValue":
                raise _ValueError(
                    "Fn::FindInMap parameter only supports 'DefaultValue'", value
                )
            self._value = _ForEachValue.create(v)

    def value(
        self, cfn, params: Optional[Mapping[str, Any]] = None, only_params: bool = False
    ):
        if params is None:
            params = {}

        return self._value.value(cfn, params, only_params)


class _ForEachValueFnFindInMap(_ForEachValue):
    def __init__(self, _hash: str, obj: Any) -> None:
        super().__init__(_hash)
        if not isinstance(obj, list):
            raise _TypeError("Fn::FindInMap should be a list", obj)

        if len(obj) not in [3, 4]:
            raise _ValueError("Fn::FindInMap requires a list of 3 or 4 values", obj)

        self._map = [
            _ForEachValue.create(obj[0]),
            _ForEachValue.create(obj[1]),
            _ForEachValue.create(obj[2]),
        ]
        if len(obj) == 4:
            self._map.append(_FnFindInMapDefaultValue(get_hash(obj[3]), obj[3]))

        self._obj = obj

    def value(
        self,
        cfn: Any,
        params: Optional[Mapping[str, Any]] = None,
        only_params: bool = False,
    ) -> Any:
        if params is None:
            params = {}
        t_map = deepcopy(self._map)
        mapping = None
        try:
            mapping = cfn.template.get("Mappings", {}).get(
                t_map[0].value(cfn, params, only_params)
            )
        except Exception:  # pylint: disable=broad-exception-caught
            if len(cfn.template.get("Mappings", {}).keys()) == 1:
                mapping = cfn.template.get("Mappings", {}).get(
                    list(cfn.template.get("Mappings", {}).keys())[0]
                )

        try:
            if mapping is None and isinstance(
                t_map[1].value(cfn, params, only_params), str
            ):
                for k, v in cfn.template.get("Mappings", {}).items():
                    if isinstance(v, dict):
                        if t_map[1].value(cfn) in v:
                            t_map[0] = _ForEachValue.create(k)
                            mapping = v
                            break
        except _ResolveError:
            pass

        try:
            if mapping is None and isinstance(
                t_map[2].value(cfn, params, only_params), str
            ):
                for m1, mv1 in cfn.template.get("Mappings", {}).items():
                    if isinstance(mv1, dict):
                        for k1, kv1 in mv1.items():
                            if isinstance(kv1, dict):
                                if t_map[2].value(cfn, params, only_params) in kv1:
                                    t_map[1] = _ForEachValue.create(k1)
                                    t_map[0] = _ForEachValue.create(m1)
                                    mapping = mv1
                                    break
        except _ResolveError:
            pass

        if mapping:
            try:
                t_map[1].value(cfn)
            except _ResolveError:
                try:
                    t_map[2].value(cfn)
                    for k, v in mapping.items():
                        if isinstance(v, dict):
                            if t_map[2].value(cfn, params, only_params) in v:
                                t_map[1] = _ForEachValue.create(k)
                except _ResolveError:
                    pass

        if mapping:
            try:
                return mapping.get(t_map[1].value(cfn, params, only_params), {}).get(
                    t_map[2].value(cfn, params, only_params)
                )
            except _ResolveError as e:
                if len(self._map) == 4:
                    return self._map[3].value(cfn, params, only_params)
                raise _ResolveError("Can't resolve Fn::FindInMap", self._obj) from e

        if len(self._map) == 4:
            return self._map[3].value(cfn, params, only_params)
        raise _ResolveError("Can't resolve Fn::FindInMap", self._obj)


class _ForEachValueRef(_ForEachValue):
    def __init__(self, _hash: str, obj: Any) -> None:
        super().__init__(_hash)
        if not isinstance(obj, (str, dict)):
            raise _TypeError("Fn::FindInMap should be a list", obj)

        self._ref = _ForEachValue.create(obj)
        self._obj = obj

    # pylint: disable=too-many-return-statements
    def value(
        self,
        cfn: Any,
        params: Optional[Mapping[str, Any]] = None,
        only_params: bool = False,
    ) -> Any:
        if params is None:
            params = {}
        v = self._ref.value(cfn)

        if not isinstance(v, str):
            raise _ResolveError("Can't resolve Fn::Ref", self._obj)

        if v in params:
            return params[v]

        if only_params:
            raise _ResolveError("Can't resolve Fn::Ref", self._obj)

        region = cfn.regions[0]
        account_id = "123456789012"
        partition = "aws"
        if region in ("us-gov-east-1", "us-gov-west-1"):
            partition = "aws-us-gov"
        if region in ("cn-north-1", "cn-northwest-1"):
            partition = "aws-cn"
        if v == "AWS::Region":
            return region

        if v == "AWS::AccountId":
            return account_id

        if v == "AWS::NotificationARNs":
            return [f"arn:{partition}:sns:{region}:{account_id}:notification"]

        if v == "AWS::NoValue":
            return None

        if v == "AWS::Partition":
            return partition

        if v == "AWS::StackId":
            return (
                f"arn:{partition}:cloudformation:"
                f"{region}:{account_id}:"
                "stack/teststack/51af3dc0-da77-11e4-872e-1234567db123"
            )

        if v == "AWS::StackName":
            return "teststack"

        if v == "AWS::URLSuffix":
            if region in ("cn-north-1", "cn-northwest-1"):
                return "amazonaws.com.cn"

            return "amazonaws.com"

        p = cfn.template.get("Parameters", {}).get(v)
        if not p:
            raise _ResolveError("Can't resolve Fn::Ref", self._obj)
        t = p.get("Type", "String")
        default = p.get("Default")
        if default:
            if "List" in t:
                return [x.strip() for x in default.split(",")]
            return default
        allowed_values = p.get("AllowedValues", [])
        if len(allowed_values) > 0:
            if "List" in t:
                return [x.strip() for x in allowed_values[0].split(",")]
            return allowed_values[0]

        raise _ResolveError("Can't resolve Fn::Ref", self._obj)


class _ForEachCollection:
    def __init__(self, obj: Any) -> None:
        self._collection: Optional[List[_ForEachValue]] = None
        self._obj = obj
        self._fn: Optional[_ForEachValue] = None
        if isinstance(obj, list):
            self._collection = []
            self._string = obj
            for item in obj:
                self._collection.append(_ForEachValue.create(item))
            return
        if isinstance(obj, dict):
            self._fn = _ForEachValue.create(obj)
            return
        raise _TypeError("Collection must be a list or an object", obj)

    def values(
        self, cfn: Any, collection_cache: MutableMapping[str, str]
    ) -> Iterable[str]:
        if self._collection:
            for item in self._collection:
                try:
                    yield item.value(cfn, {}, False)
                except _ResolveError:
                    v = "".join(random.choices(string.ascii_letters, k=_N))  # nosec
                    collection_cache[item.hash] = v
                    yield v
            return
        if self._fn:
            try:
                values = self._fn.value(cfn, {}, False)
                if values:
                    if isinstance(values, list):
                        for value in values:
                            if isinstance(value, str):
                                yield value
                            else:
                                raise _ValueError(
                                    f"Fn::ForEach collection value must be a {_SCALAR_TYPES!r}",
                                    self._obj,
                                )
                        return
                    raise _ValueError(
                        "Fn::ForEach collection must return a list", self._obj
                    )
            except _ResolveError:
                if self._fn.hash in collection_cache:
                    yield from iter(collection_cache[self._fn.hash])
                else:
                    collection_cache[self._fn.hash] = []
                    for _ in range(0, 2):
                        v = "".join(random.choices(string.ascii_letters, k=_N))  # nosec
                        collection_cache[self._fn.hash].append(v)
                        yield v
                return
        raise _ResolveError("Fn::ForEach could not be resolved", self._obj)


class _ForEachOutput:
    def __init__(self, obj: Any) -> None:
        if isinstance(obj, dict):
            self._output = obj
            return
        raise _TypeError("Output must be a dict", obj)

    def value(self) -> Any:
        return self._output


class _ForEach:
    def __init__(
        self, key: str, value: Any, collection_cache: MutableMapping[str, str]
    ) -> None:
        self._key = key
        self._collection_cache: MutableMapping[str, str] = collection_cache
        if not isinstance(value, list):
            raise _TypeError("Fn::ForEach values must be a list of 3 elements", key)

        if not len(value) == 3:
            raise _TypeError("Fn::ForEach values must be a list of 3 elements", key)

        self._identifier = _ForEachValue.create(value[0])
        self._collection = _ForEachCollection(value[1])
        self._output = _ForEachOutput(value[2])

    def items(self, cfn: Any) -> Iterable[Tuple[str, str]]:
        items = self._collection.values(cfn, self._collection_cache)
        yield from iter(items)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Select(CloudFormationLintRule):
    """Check if Select values are correct"""

    id = "E1017"
    shortdesc = "Select validation of parameters"
    description = "Making sure the Select function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-select.html"
    tags = ["functions", "select"]
    supported_functions = [
        "Fn::FindInMap",
        "Fn::GetAtt",
        "Fn::GetAZs",
        "Fn::If",
        "Fn::Split",
        "Fn::Cidr",
        "Fn::Select",  # issue: 2895
        "Ref",
    ]

    def _test_index_obj(self, index_obj, path):
        matches = []
        if isinstance(index_obj, dict):
            if len(index_obj) == 1:
                for index_key, _ in index_obj.items():
                    if index_key not in [
                        "Ref",
                        "Fn::FindInMap",
                        "Fn::Select",
                    ]:
                        message = "Select index should be an Integer or a function Ref, Fn::FindInMap, or Fn::Select for {0}"
                        matches.append(
                            RuleMatch(
                                path,
                                message.format("/".join(map(str, path))),
                            )
                        )
            else:
                message = "Select index should be an Integer or a function Ref, Fn::FindInMap, or Fn::Select for {0}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format("/".join(map(str, path))),
                    )
                )
        elif not isinstance(index_obj, int):
            try:
                int(index_obj)
            except (ValueError, TypeError):
                message = "Select index should be an Integer or a function of Ref, Fn::FindInMap, or Fn::Select for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )

        return matches

    def _test_list_obj(self, list_obj, path):
        matches = []
        if isinstance(list_obj, dict):
            if len(list_obj) == 1:
                for key, _ in list_obj.items():
                    if key not in self.supported_functions:
                        message = "Select should use a supported function of {0}"
                        matches.append(
                            RuleMatch(
                                path,
                                message.format(
                                    ", ".join(map(str, self.supported_functions))
                                ),
                            )
                        )
            else:
                message = "Select should use a supported function of {0}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(", ".join(map(str, self.supported_functions))),
                    )
                )
        elif not isinstance(list_obj, list):
            message = "Select should be an array of values for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))

        return matches

    def _test_select_obj(self, select_obj, path):
        matches = []
        if not isinstance(select_obj, list):
            message = "Select should be a list of 2 elements for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
            return matches
        if len(select_obj) != 2:
            message = "Select should be a list of 2 elements for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
            return matches

        index_obj = select_obj[0]
        list_of_objs = select_obj[1]
        matches.extend(self._test_index_obj(index_obj, path[:] + [0]))
        matches.extend(self._test_list_obj(list_of_objs, path[:] + [1]))

        return matches

    def match(self, cfn):
        matches = []

        select_objs = cfn.search_deep_keys("Fn::Select")

        for select_obj in select_objs:
            select_value_obj = select_obj[-1]
            tree = select_obj[:-1]
            matches.extend(self._test_select_obj(select_value_obj, tree[:]))

        return matches
